// CoverIt - coverage analysis plugin for IDA v5.0
// Copyright 2006 (c) Ilfak Guilfanov <ig@hexblog.com>
// Freeware. Please keep the copyright message intact.
//
// This plugin runs the debugged application and colors the executed
// instructions with the specified color.
//
// The instruction that did not execute will not change the color.
// If you want to clear the instruction colors, select the whole
// code segment and use the 'Other, Set insn color' command.
// Choose an undefined custom color and all instruction colors will
// be reset. Another option is to modify this plugin to reset
// colors at each start. I decided not to do it to allow incremental
// instruction colorizing.
//
// Possible modifications and improvements of the plugin:
//      - Track execution of functions instead of basic blocks.
//      - Create a list of executed/not-executed functions.
//      - Create something like a navigation band to display the results
//        (in fact it is not very difficult, just create a window and draw on it
//         pixel by pixel or, rather, rectangle by rectangle)
//      - Count the number of executions. Currently the plugin detects only
//        the fact of the instruction execution but does not count how many
//        times it gets executed. Counting will slow things down but I'm
//        sure that it can still be made acceptably fast.
//      - Monitor several segments/dlls at once. The current version handles
//        only the first code segment coming from the input file (so called
//        loader segment). It can be made to monitor the whole memory
//        process excluding the windows kernel code that handles exceptions
//      - Port to other platforms and processors. For the monet the code is
//        MS Windows-oriented (the exception and breakpoint codes are hardcoded).
//        Seems to be easy.
//      - Make the plugin to remember the basic block list between runs.
//        This will improve the startup speed of subsequent runs.
//      - Add customization dialog box (the color, function/bb selector),
//        in short, everything said above can be parameterized.
//
// Have fun!
//
// This plugin can be used as a nice test tool, btw :)

#include <vector>
#include <map>

#include <ida.hpp>
#include <idp.hpp>
#include <dbg.hpp>
#include <auto.hpp>
#include <bytes.hpp>
#include <loader.hpp>
#include <kernwin.hpp>

//--------------------------------------------------------------------------
// Our callback declarations
typedef int     idaapi get_debug_event_t(debug_event_t *event, bool ida_is_idle);
typedef ssize_t idaapi read_memory_t(ea_t ea, void *buf, size_t size);
typedef ssize_t idaapi write_memory_t(ea_t ea, const void *buf, size_t size);

static get_debug_event_t my_get_debug_event;
static read_memory_t     my_read_memory;
static write_memory_t    my_write_memory;

// Pointers to old callbacks
static get_debug_event_t *old_get_debug_event = my_get_debug_event;
static read_memory_t     *old_read_memory     = my_read_memory;
static write_memory_t    *old_write_memory    = my_write_memory;

//--------------------------------------------------------------------------
// somewhat greenish color
static const bgcolor_t executed_insn_color = 0xA0FFA0;

typedef std::vector<bool> progbits_t;
typedef std::vector<uchar> progbytes_t;
typedef std::map<ea_t, asize_t> basic_blocks_t;

static ea_t base;          // current instrumentation base address
static basic_blocks_t bbs; // basic blocks
static progbits_t bpts;    // addresses of bpts embedded into the program
static progbytes_t core;   // copy of the program code segment
static int ip_idx;         // index of ip register

// pointer to the UI refresher function
// Used to refresh colorized instructions on the screen
void (idaapi*idaview_marker)(ea_t ea);
//--------------------------------------------------------------------------
// Colorize a block of instructions
static void colorize_block(ea_t ea, asize_t size, bgcolor_t color)
{
  // tell ida to refresh the basic block next time
  mark_idaview_for_refresh(ea);
  ea_t end = ea + size;
  while ( ea < end )
  {
    set_item_color(ea, color);
    ea = nextthat(ea, end, f_isCode, NULL);
  }
}

//--------------------------------------------------------------------------
// Cleanup breakpoint and core information
static void cleanup(void)
{
  bbs.clear();
  bpts.clear();
  core.clear();
}

//--------------------------------------------------------------------------
// Find all basic blocks in the specified area
static bool gather_basic_blocks(ea_t ea1, ea_t ea2)
{
  show_wait_box("Finding basic blocks");
  ea_t start = BADADDR;
  bool ok = true;
  int cnt = 0;
  while ( true )
  {
    if ( wasBreak() )
    {
      ok = false;
      break;
    }
    if ( ++cnt == 1000 )
    {
      showAddr(ea1);
      cnt = 0;
    }
    if ( start == BADADDR )
    {
      ea1 = nextthat(ea1, ea2, f_isCode, NULL);
      if ( ea1 >= ea2 )
        break;
      start = ea1;
    }
    while ( ea1 < ea2 )
    {
      if ( !ua_ana0(ea1) )
        break;
      ea1 = get_item_end(ea1);
      if ( is_basic_block_end(true) )
        break;
    }
    if ( ea1 != start )
      bbs[start] = ea1 - start; // remember the bb start and size
    if ( !isCode(get_flags_novalue(ea1)) )
      start = BADADDR;
    else
      start = ea1;
  }
  hide_wait_box();
  return ok;
}

//--------------------------------------------------------------------------
// Create a breakpoint for each basic block start
// Since we are going to use each breakpoint only once, we will
// modify the process memory directly instead of using the
// breakpoint mechanism of IDA Pro. This way the breakpoints
// can be handled much faster
static bool add_bpts(void)
{
  progbytes_t c2 = core; // make a copy of the code segment
  for ( basic_blocks_t::iterator p=bbs.begin(); p != bbs.end(); ++p )
  {
    size_t i = p->first - base;
    c2[i] = 0xCC;
    bpts[i] = true;     // remember the bpt address
  }
  // Write all breakpoints at once (update the whole code segment)
  return dbg->write_memory(base, &c2[0], c2.size()) > 0;
}

//--------------------------------------------------------------------------
// Create all necessary data structures to perform coverage analysis of
// the specified area.
static bool instrument_area(ea_t ea1, ea_t ea2)
{
  base = ea1;
  asize_t size = ea2 - ea1;
  bpts.resize(size);
  core.resize(size);
  if ( dbg->read_memory(base, &core[0], size) <= 0 )
  {
    warning("Could not read process memory");
    return false;
  }
  if ( !gather_basic_blocks(ea1, ea2) )
    return false;
  if ( !add_bpts() )
  {
    warning("Could not write process memory");
    return false;
  }
  return true;
}

//--------------------------------------------------------------------------
// Create all necessary data structures to perform coverage analysis of
// the program.
static bool create_instrumenting_bpts(void)
{
  // we will instrument the first loader segment of 'SEG_CODE' type
  // Usually it is the segment containing the most interesting code
  int nseg = get_segm_qty();
  for ( int i=0; i < nseg; i++ )
  {
    segment_t *s = getnseg(i);
    if ( s == NULL )
      continue;
    if ( !s->is_loader_segm() )
      continue;
    if ( s->type != SEG_CODE )
      continue;
    // found it
    return instrument_area(s->startEA, s->endEA);
  }
  warning("Could not find a segment to analyze.\n"
          "The plugin failed to locate a loader code segment.");
  return false;
}

//--------------------------------------------------------------------------
// Got an exception caused by a software breakpoint
// Remove the breakpoint, and reexecute the instruction
static bool restore_program_byte(const debug_event_t *event)
{
  // check the address range
  if ( event->ea < base )
    return false;
  size_t i = event->ea - base;
  if ( i >= core.size() )
    return false;
  if ( !bpts[i] )  // do we still have a breakpoint there?
    return false;

  // yes, this is an exception caused by our breakpoint
  // remove it
  if ( old_write_memory(event->ea, &core[i], 1) <= 0 )
    return false;
  bpts[i] = false;

  // step back
  regval_t rv;
  rv.ival = event->ea;
  if ( dbg->thread_write_register(event->tid, ip_idx, &rv) <= 0 )
    return false;

  // colorize the basic block on the fly
  // this way the user will see the program coverage dynamics
  // as the program runs
  colorize_block(event->ea, bbs[event->ea], executed_insn_color);
  return true;
}

//--------------------------------------------------------------------------
// Callback to call at each debug event
static int idaapi my_get_debug_event(debug_event_t *event, bool ida_is_idle)
{
  int code;
  while ( true )
  {
    code = old_get_debug_event(event, ida_is_idle);
    if ( code != 1 )
      break;
    // check if this is an exception caused by our breakpoint
    if ( event->eid != EXCEPTION )
      break;
    if ( event->exc.code != 0x80000003 ) // breakpoint?
      break;
    if ( !restore_program_byte(event) )
      break;
    // yes, it was our breakpoint, we took it into account and colorized bb
    // now continue the execution and do not report this
    // debug event to ida.
    if ( dbg->continue_after_event(event) <= 0 )
    {
      // something really wrong
      warning("%a: could not continue after breakpoint", event->ea);
      break;
    }
    // make it fast now, do not wait for events if there is none
    ida_is_idle = false;
  }
  return code;
}

//--------------------------------------------------------------------------
// Read memory, replace all our bpts by original bytes
// This function hides our bpts from the rest of ida
static ssize_t idaapi my_read_memory(ea_t ea, void *buf, size_t size)
{
  int code = old_read_memory(ea, buf, size);
  if ( code > 0 )
  {
    // replace 0xCCs by the original values
    uchar *ptr = (uchar *)buf;
    adiff_t i1 = ea - base;
    adiff_t i2 = i1 + size;
    if ( i2 > 0 )
    {
      if ( i2 > core.size() )
        i2 = core.size();
      adiff_t n = i2 - i1;
      for ( adiff_t i=qmax(-i1, 0); i < n; i++ )
      {
        if ( bpts[i1+i] && ptr[i] == 0xCC )
          ptr[i] = core[i1+i];
      }
    }
  }
  return code;
}

//--------------------------------------------------------------------------
// Write process memory
// Since the rest of IDA is not aware of our breakpoints, we
// have to restore them at the write to the process memory
static ssize_t idaapi my_write_memory(ea_t ea, const void *buf, size_t size)
{
  if ( ea >= base+core.size() || ea+size <= base || size == 0 )
    return old_write_memory(ea, buf, size);

  uchar *b2 = (uchar *)qalloc(size);
  if ( b2 == NULL )
    nomem("my_write_memory");
  memcpy(b2, buf, size);

  // replace original value by 0xCCs
  uchar *ptr = (uchar *)b2;
  adiff_t i1 = ea - base;
  adiff_t i2 = i1 + size;
  if ( i2 > core.size() )
    i2 = core.size();
  adiff_t n = i2 - i1;
  for ( adiff_t i=qmax(-i1, 0); i < n; i++ )
  {
    if ( bpts[i1+i] )
      ptr[i] = 0xCC;
  }
  int code = old_write_memory(ea, b2, size);
  qfree(b2);
  return code;
}

//--------------------------------------------------------------------------
// Set or remove our callbacks for debugger module functions
// This way we stay between the ida kernel and the debugger module.
// The IDA kernel is not aware of our breakpoints.
// The debugger module considers them as simple memory patches.
bool switch_dbg_functions(void)
{
  if ( dbg != NULL )
  {
    qswap(old_get_debug_event, dbg->get_debug_event);
    qswap(old_read_memory, dbg->read_memory);
    qswap(old_write_memory, dbg->write_memory);
    return true;
  }
  return false;
}

//--------------------------------------------------------------------------
// The debugger event callback
// It is used to start and stop the instrumentation mechanism
static int idaapi callback(void * /*user_data*/, int notification_code, va_list va)
{
  static int stage;
  static char needed_file[QMAXPATH];

  switch ( notification_code )
  {
    case dbg_process_start:
    case dbg_process_attach:
      get_input_file_path(needed_file, sizeof(needed_file));
      // no break
    case dbg_library_load:
      if ( stage == 0 )
      {
        const debug_event_t *pev = va_arg(va, const debug_event_t *);
        if ( stricmp(pev->modinfo.name, needed_file) != 0 )
          break;
        if ( create_instrumenting_bpts() )
        {
          switch_dbg_functions();
          stage++;
        }
      }
      break;

    case dbg_process_exit:
      if ( stage )
      {
        stage = 0;
        cleanup();
        switch_dbg_functions();
        unhook_from_notification_point(HT_DBG, callback, NULL);
      }
      break;
  }
  return 0;
}

//--------------------------------------------------------------------------
void run(int /*arg*/)
{
  // can not work without the debugger
  if ( dbg == NULL )
  {
    warning("The debugger is missing");
    return;
  }

  // find ip register index - will need it later
  ip_idx = -1;
  for ( int i=0; i < dbg->registers_size; i++ )
  {
    if ( (dbg->registers[i].flags & REGISTER_IP) != 0 )
    {
      ip_idx = i;
      break;
    }
  }
  if ( ip_idx == -1 )
  {
    warning("interr: could not find ip register");
    return;
  }

  // The analysis has to be finished
  // Otherwise the basic block information will be incomplete
  if ( !autoIsOk() )
  {
    if ( askyn_c(0, "HIDECANCEL\n"
                    "The analysis has not finished yet.\n"
                    "The plugin will give wrong results.\n"
                    "Do you want to continue?") <= 0 )
      return;
  }

  // Ask for the confirmation since we are going to run user code
  static const char str[] =
  "HIDECANCEL\n"
  "AUTOHIDE REGISTRY\n"
  "Coverage analysis plugin\n"
  "\n"
  "This plugin will run the disassembled program.\n"
  "\n"
  "Do you really want to launch the program?\n";

  if ( askyn_c(0, str) <= 0 )
    return;

  // Hook to detect the process start
  if ( !hook_to_notification_point(HT_DBG, callback, NULL) )
  {
    warning("Could not hook to notification point\n");
    return;
  }

  // Now let's start the debugger
  // The rest of work will be done in the callbacks
  if ( !start_process() )
  {
    warning("Sorry, could not start the process");
    unhook_from_notification_point(HT_DBG, callback, NULL);
  }
}

//--------------------------------------------------------------------------
int init(void)
{
  // Our plugin works only for x86 PE executables
  if ( ph.id != PLFM_386 || inf.filetype != f_PE )
    return PLUGIN_SKIP;
  // Get a pointer to the function to refresh the disassembly
  setup_idaview_marker();
  return PLUGIN_OK;
}

//--------------------------------------------------------------------------
void term(void)
{
  // just to be safe, normally this is not required
  unhook_from_notification_point(HT_DBG, callback, NULL);
}

//--------------------------------------------------------------------------
char wanted_name[] = "Cover It!";
char wanted_hotkey[] = "";


//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------
plugin_t PLUGIN =
{
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,                 // initialize

  term,                 // terminate. this pointer may be NULL.

  run,                  // invoke plugin

  wanted_name,          // long comment about the plugin
                        // it could appear in the status line
                        // or as a hint

  wanted_name,          // multiline help about the plugin

  wanted_name,          // the preferred short name of the plugin
  wanted_hotkey         // the preferred hotkey to run the plugin
};
